#!/usr/bin/env python

import sys

from dark.errors import ReadLengthsNotIdenticalError
from dark.reads import (
    addFASTACommandLineOptions, parseFASTACommandLineOptions, Reads)


def printHeader(variableSites, args, baseOffset):
    n = len(variableSites)
    print('%d site%s %svariable (threshold for homogeneity: %.3f).' %
          (n, ' was' if n == 1 else 's were',
           'confirmed ' if args.confirm else '', args.homogeneous),
          file=sys.stderr)
    if args.reference:
        print('Offsets adjusted by %d, relative to sequence %r.' %
              (baseOffset, args.reference), file=sys.stderr)


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            'Given FASTA on standard input, write reads with only the '
            'variable sites to standard output. If no sites are variable, '
            'nothing is written to standard output.'))

    parser.add_argument(
        '--printSites', action='store_true', default=False,
        help=('Print the variable site locations to standard error along '
              'with the nucleotides found at the site, the homogeneity '
              'fraction of the site and the reference base (if a reference '
              'is given).'))

    parser.add_argument(
        '--sitesOnly', action='store_true', default=False,
        help=('Only print the (comma separated) variable sites to standard '
              'output. This output can then be used directly as an argument '
              'for --keepSites to filter-fasta.py'))

    parser.add_argument(
        '--homogeneous', default=1.0, type=float,
        help=('If the most-common nucleotide frequency at a site is at least '
              'this value, the site will be considered homogeneous.'))

    parser.add_argument(
        '--confirm', action='store_true', default=False,
        help=('Only keep sites where there is confirm variation (i.e., '
              'ambiguous sites that are compatible with there being no '
              'variation are not included).'))

    parser.add_argument(
        '--reference',
        help=('A sequence id. Site offsets will be reported relative to this '
              'sequence. That means the number of gap (hyphen) chars at the '
              'start of this sequence will be counted and offsets will be '
              'incremented by that amount (and lower offsets in sequences '
              'that start before the specified sequence will be ignored).'))

    parser.add_argument(
        '--unknownAreAmbiguous', action='store_true', default=False,
        help=("Any unknown character (e.g., a '-' gap or '?' unknown base) "
              "will be treated as being fully ambiguous (i.e., could be any "
              "of ACGT). Otherwise, all unknown characters are counted "
              "as '-' characters."))

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = Reads(list(parseFASTACommandLineOptions(args)))

    if not reads:
        sys.exit(0)

    if args.reference:
        for read in reads:
            if read.id == args.reference:
                break
        else:
            print('Could not find --reference sequence %r.' % args.reference,
                  file=sys.stderr)
            sys.exit(1)
        baseOffset = len(read.sequence) - len(read.sequence.lstrip('-'))
        reference = read
    else:
        baseOffset = 0
        reference = None

    try:
        variableSites = reads.variableSites(
            confirm=args.confirm, homogeneityLevel=args.homogeneous,
            unknownAreAmbiguous=args.unknownAreAmbiguous)
    except ReadLengthsNotIdenticalError:
        print('Input sequences are not all the same length!', file=sys.stderr)
        sys.exit(2)

    if variableSites:
        toDelete = set()
        if args.printSites:
            for site, counts in variableSites.items():
                if site >= baseOffset:
                    ref = ((' (ref %s)' % reference.sequence[site])
                           if reference else '')
                    print('%d: %s%s' % (site + 1 - baseOffset, counts, ref),
                          file=sys.stderr)
                else:
                    toDelete.add(site)

        for site in toDelete:
            del variableSites[site]

    if variableSites:
        if args.sitesOnly:
            print(','.join(map(lambda site: str(site + 1 - baseOffset),
                               sorted(variableSites))))
        else:
            saveAs = 'fasta' if args.fasta else 'fastq'
            reads.filter(keepSites=set(variableSites)).save(sys.stdout, saveAs)
            printHeader(variableSites, args, baseOffset)
    else:
        print('No sites were %svariable (threshold for homogeneity: %.3f).' %
              ('confirmed ' if args.confirm else '', args.homogeneous),
              file=sys.stderr)



from unittest import TestCase
from six import assertRaisesRegex

from dark.dimension import dimensionalIterator


class TestDimensionalIterator(TestCase):
    """
    Test the dimensionalIterator generator.
    """

    def testNoDimensions(self):
        self.assertEqual(
            (),
            tuple(dimensionalIterator(())))

    def testZeroDimension(self):
        """
        Passing a zero dimension must result in a ValueError with the expected
        error string.
        """
        error = r'^Dimensions not all positive! \(2, 0, 3\)$'
        assertRaisesRegex(self, ValueError, error, next,
                          dimensionalIterator((2, 0, 3)))

    def testNegativeDimension(self):
        """
        Passing a negative dimension must result in a ValueError with the
        expected error string.
        """
        error = r'^Dimensions not all positive! \(2, -1, 3\)$'
        assertRaisesRegex(self, ValueError, error, next,
                          dimensionalIterator((2, -1, 3)))

    def testLimitedTo0Items(self):
        self.assertEqual(
            (),
            tuple(dimensionalIterator((2, 2), maxItems=0)))

    def test1X2(self):
        self.assertEqual(
            ((0, 0), (0, 1)),
            tuple(dimensionalIterator((1, 2))))

    def test2X2(self):
        self.assertEqual(
            ((0, 0), (0, 1),
             (1, 0), (1, 1)),
            tuple(dimensionalIterator((2, 2))))

    def test2X2LimitedTo4Items(self):
        """
        If the passed maximum number of items is the same as the number
        of items we'd expect to get with no limit, we should get all
        items.
        """
        self.assertEqual(
            ((0, 0), (0, 1),
             (1, 0), (1, 1)),
            tuple(dimensionalIterator((2, 2), maxItems=4)))

    def test2X2LimitedTo3Items(self):
        self.assertEqual(
            ((0, 0), (0, 1),
             (1, 0)),
            tuple(dimensionalIterator((2, 2), maxItems=3)))

    def testStarX2LimitedTo10Items(self):
        self.assertEqual(
            ((0, 0), (0, 1),
             (1, 0), (1, 1),
             (2, 0), (2, 1),
             (3, 0), (3, 1),
             (4, 0), (4, 1)),
            tuple(dimensionalIterator(('*', 2), maxItems=10)))

    def testStarX3LimitedTo10Items(self):
        self.assertEqual(
            ((0, 0), (0, 1), (0, 2),
             (1, 0), (1, 1), (1, 2),
             (2, 0), (2, 1), (2, 2),
             (3, 0)),
            tuple(dimensionalIterator(('*', 3), maxItems=10)))

    def test3X2XStarX4(self):
        """
        If '*' is used in a dimension that's not the first, iteration
        should get 'stuck' there. I.e., the earlier dimensions will always
        return zero as the '*' dimension is never exhauted.
        """
        self.assertEqual(
            ((0, 0, 0, 0), (0, 0, 0, 1), (0, 0, 0, 2), (0, 0, 0, 3),
             (0, 0, 1, 0), (0, 0, 1, 1), (0, 0, 1, 2), (0, 0, 1, 3),
             (0, 0, 2, 0), (0, 0, 2, 1), (0, 0, 2, 2), (0, 0, 2, 3),
             (0, 0, 3, 0), (0, 0, 3, 1), (0, 0, 3, 2), (0, 0, 3, 3),
             (0, 0, 4, 0), (0, 0, 4, 1), (0, 0, 4, 2), (0, 0, 4, 3)),
            tuple(dimensionalIterator((3, 2, '*', 4), maxItems=20)))

    def test2X3X4(self):
        self.assertEqual(
            ((0, 0, 0), (0, 0, 1), (0, 0, 2), (0, 0, 3),
             (0, 1, 0), (0, 1, 1), (0, 1, 2), (0, 1, 3),
             (0, 2, 0), (0, 2, 1), (0, 2, 2), (0, 2, 3),
             (1, 0, 0), (1, 0, 1), (1, 0, 2), (1, 0, 3),
             (1, 1, 0), (1, 1, 1), (1, 1, 2), (1, 1, 3),
             (1, 2, 0), (1, 2, 1), (1, 2, 2), (1, 2, 3)),
            tuple(dimensionalIterator((2, 3, 4))))

    def test1X1X1(self):
        self.assertEqual(
            ((0, 0, 0),),
            tuple(dimensionalIterator((1, 1, 1))))



#!/usr/bin/env python

from __future__ import print_function

import sys
import argparse
from collections import Counter
from numpy import std

from dark.filter import (
    addFASTAFilteringCommandLineOptions, parseFASTAFilteringCommandLineOptions)
from dark.reads import Reads
from dark.sam import samfile, SAMFilter, samReferences, UnknownReference
from dark.utils import baseCountsToStr, pct


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Print SAM/BAM file coverage statistics by offset. '
                 'Output lines show the offset.'))

addFASTAFilteringCommandLineOptions(parser)
SAMFilter.addFilteringOptions(parser, samfileIsPositional=True)

parser.add_argument(
    '--noOffsets', default=False, action='store_true',
    help='Do not print per-offset details of base counts.')

parser.add_argument(
    '--noStats', default=False, action='store_true',
    help='Do not print final average and standard deviation statistics.')

parser.add_argument(
    '--noFilter', default=False, action='store_true',
    help=('Do not use our SAM filtering. Note that if you give this option, '
          'any filtering option (other than --referenceId) you also specify '
          'that is provided by the SAMFilter.addFilteringOptions will be '
          'silently ignored!'))

args = parser.parse_args()

if args.noOffsets and args.noStats:
    print('You have used both --noOffsets and --noStats, so there is no '
          'output!', file=sys.stderr)
    sys.exit(1)


# We don't have a file of reads, we just want a read filter that we can use
# to filter the SAM file query sequences and to get reference lengths from.
reads = parseFASTAFilteringCommandLineOptions(args, Reads())
samFilter = SAMFilter.parseFilteringOptions(args, reads.filterRead)

printOffsets = not args.noOffsets
printStats = not args.noStats

if samFilter.referenceIds and len(samFilter.referenceIds) > 1:
    print('Only one reference id can be given. To calculate coverage for more '
          'than one reference, run this script multiple times.',
          file=sys.stderr)
    sys.exit(1)

try:
    referenceLengths = samFilter.referenceLengths()
except UnknownReference:
    referenceId = samFilter.referenceIds.pop()
    referenceIds = samReferences(args.samfile)
    print('Reference %r does not appear in SAM file %s. Known '
          'references are: %s.' % (
              referenceId, args.samfile, ', '.join(sorted(referenceIds))),
          file=sys.stderr)
    sys.exit(1)


if args.noFilter:
    # Do not do our custom SAM filtering.
    def filterRead(read):
        return not (read.is_del or read.is_refskip)
else:
    def filterRead(read):
        return (not (read.is_del or read.is_refskip) and
                samFilter.filterAlignment(read.alignment))


if printStats:
    counts = []

with samfile(args.samfile) as sam:

    if samFilter.referenceIds:
        # No need to check if the given reference id is in referenceLengths
        # because the samFilter.referenceLengths call above catches that.
        referenceId = samFilter.referenceIds.pop()
    else:
        if len(referenceLengths) == 1:
            referenceId = list(referenceLengths)[0]
        else:
            print('SAM file %r contains %d references (%s). Only one '
                  'reference id can be analyzed at a time. Please use '
                  '--referenceId to specify the one you want examined.' % (
                      args.samfile, len(referenceLengths),
                      ', '.join(sorted(referenceLengths))), file=sys.stderr)
            sys.exit(1)

    for column in sam.pileup(reference=referenceId):
        bases = Counter()
        for read in column.pileups:
            if filterRead(read):
                base = read.alignment.query_sequence[read.query_position]
                bases[base] += 1

        baseCount = sum(bases.values())

        if printStats:
            counts.append(baseCount)

        if printOffsets:
            print('%d: %d %s' % (column.reference_pos + 1, baseCount,
                                 baseCountsToStr(bases)))

if printStats:
    referenceLength = referenceLengths[referenceId]
    print('Reference id: %s' % referenceId)
    print('Reference length: %d' % referenceLength)
    print('Bases covered: %s' % pct(len(counts), referenceLength))
    print('Min coverage depth: %d' % (
        0 if len(counts) < referenceLength else min(counts)))
    if counts:
        # Don't use Python3 default= option on max. Trying to keep Python 2
        # compatibility.
        print('Max coverage depth: %d' % max(counts))
    else:
        print('Max coverage depth: 0')
    print('Mean coverage depth: %.3f' % (sum(counts) / referenceLength))
    if counts:
        print('Coverage depth s.d.: %.3f' % std(counts))



def printHSP(hsp, indent=''):
    for attr in ['bits', 'expect', 'frame', 'query_end', 'query_start',
                 'sbjct', 'query', 'sbjct_end', 'sbjct_start']:
        print('%s%s: %s' % (indent, attr, hsp[attr]))


def normalizeHSP(hsp, readLen, blastApplication):
    """
    Examine an HSP and return information about where the query and subject
    match begins and ends.  Return a dict with keys that allow the query to
    be displayed against the subject. The returned readStartInSubject and
    readEndInSubject indices are offsets into the subject. I.e., they
    indicate where in the subject the query falls.

    In the returned object, all indices are suitable for Python string
    slicing etc.  We must be careful to convert from the 1-based offsets
    found in BLAST output properly.

    hsp['frame'] is a (query, subject) 2-tuple, with both values coming from
    {-3, -2, -1, 1, 2, 3}. The sign indicates negative or positive sense
    (i.e., the direction of reading through the query or subject to get the
    alignment). The value is the nucleotide match offset modulo 3, plus one
    (i.e., it tells us which of the 3 possible reading frames is used in
    the match). The value is redundant because that information could also
    be obtained from the mod 3 value of the match offset.

    NOTE: the returned readStartInSubject value may be negative.  We consider
    the hit sequence to start at offset 0.  So if the read string has
    sufficient additional nucleotides before the start of the alignment
    match, it may protrude to the left of the hit. Similarly, the returned
    readEndInSubject can be greater than the subjectEnd.

    @param hsp: an HSP in the form of a C{dict}, built from a BLAST record.
        All passed hsp offsets are 1-based.
    @param readLen: the length of the read sequence.
    @param blastApplication: The C{str} command line program that was
        run (e.g., 'blastn', 'blastx').

    """

    def debugPrint(locals, msg=None):
        """
        Print debugging information showing the local variables from
        a call to normalizeHSP and then raise an C{AssertionError}.

        @param locals: A C{dict} of local variables.
        @param msg: A C{str} message to raise C{AssertionError} with.
        """
        print('normalizeHSP error:')
        print('  readLen: %d' % readLen)
        for var in sorted(locals.keys()):
            if var in ('debugPrint', 'hsp'):
                continue
            print('  %s: %s' % (var, locals[var]))
        print('  Original HSP:')
        printHSP(hsp, '    ')
        if msg:
            raise AssertionError(msg)
        else:
            raise AssertionError()

    readPositive = hsp['frame'][0] > 0
    hitPositive = hsp['frame'][1] > 0

    # The following variable names with underscores match the names of
    # attributes BioPython uses and the values (1-based) match those
    # reported by BLAST.
    read_start = hsp['query_start']
    read_end = hsp['query_end']
    sbjct_start = hsp['sbjct_start']
    sbjct_end = hsp['sbjct_end']

    # When the read is positive, BLASTN and TBLASTX give read offsets
    # ascending.
    #
    # TBLASTX reports negative read sense with indices ascending.
    # BLASTN does not report negative read sense.
    #
    # In all cases the read offsets should be ascending.
    if read_start > read_end:
        debugPrint(locals(),
                   'Assertion "read_start <= read_end" failed. Read '
                   'positive is %s. read_start = %d, read_end = %d' %
                   (readPositive, read_start, read_end))

    if hitPositive:
        # Make sure indices are ascending.
        if sbjct_start > sbjct_end:
            debugPrint(locals())
    else:
        # Hit is negative. Its indices will be ascending for TBLASTX
        # output but descending for BLASTN :-( Make sure we have them
        # ascending.
        if sbjct_start > sbjct_end:
            sbjct_start, sbjct_end = sbjct_end, sbjct_start

    # Now that we have asserted what we can about the original HSP values
    # and gotten them into ascending order, make some sane 0-based offsets.
    readStartInSubject = read_start - 1
    readEndInSubject = read_end
    subjectStart = sbjct_start - 1
    subjectEnd = sbjct_end

    if blastApplication == 'blastx':
        # In Blastx output, hit offsets are based on protein sequence
        # length but queries (and the reported offsets) are nucleotide.
        # Convert the read offsets to protein because we will plot against
        # the hit (protein).
        #
        # Note that readStartInSubject and readEndInSubject may not be 0 mod
        # 3. They are offsets into the read string giving the position of
        # the AA, which depends on the translation frame.
        readStartInSubject = int(readStartInSubject / 3)
        readEndInSubject = int(readEndInSubject / 3)

    # No operations on original 1-based HSP variables (with underscores)
    # should appear beyond this point.

    subjectLength = subjectEnd - subjectStart
    readLength = readEndInSubject - readStartInSubject

    # NOTE: readLength (above) is a really bad name. It's actually going to
    # hold the length of the match in the query. I don't know why
    # readEndInSubject - readStartInSubject is used (I mean why those two
    # variables are not named readEnd and readStart). Maybe someone made a
    # find and replace editing error which changed their names. Anyway, the
    # readLength variable is confusingly named because this function is
    # passed a 'readLen' argument, which does happen to be the full length
    # of the read.  This should be cleaned up. See ../diamond/hsp.py for
    # something cleaner.

    hitGaps = hsp['sbjct'].count('-')
    readGaps = hsp['query'].count('-')

    # Sanity check that the length of the matches in the hit and read
    # are identical, taking into account gaps in either (indicated by '-'
    # characters in the match sequences, as returned by BLAST).
    subjectLengthWithGaps = subjectLength + hitGaps
    readLengthWithGaps = readLength + readGaps
    if subjectLengthWithGaps != readLengthWithGaps:
        debugPrint(locals(),
                   'Including gaps, hit match length (%d) != Read match '
                   'length (%d)' % (subjectLengthWithGaps,
                                    readLengthWithGaps))

    # TODO: check the mod 3 value of the start offsets.

    # Calculate read indices. These are indices relative to the hit!

    # unmatchedReadLeft is the number of read bases that will be sticking
    # out to the left of the start of the hit in our plots.
    if readPositive:
        unmatchedReadLeft = readStartInSubject
    else:
        unmatchedReadLeft = readLen - readEndInSubject

    # Set the read offsets based on the direction the match with the
    # hit takes.
    if hitPositive:
        readStartInSubject = subjectStart - unmatchedReadLeft
        readEndInSubject = readStartInSubject + readLen + readGaps
    else:
        readEndInSubject = subjectEnd + unmatchedReadLeft

        readStartInSubject = readEndInSubject - readLen - readGaps

    # Final sanity checks.
    if readStartInSubject > subjectStart:
        debugPrint(locals(), 'readStartInSubject > subjectStart')
    if readEndInSubject < subjectEnd:
        debugPrint(locals(), 'readEndInSubject < subjectEnd')

    return {
        'readStart': read_start - 1,
        'readEnd': read_end,
        'readStartInSubject': readStartInSubject,
        'readEndInSubject': readEndInSubject,
        'subjectStart': subjectStart,
        'subjectEnd': subjectEnd,
    }



#!/usr/bin/env python

from __future__ import print_function

import os
import sys
import argparse
import multiprocessing
from os.path import exists, join
from tempfile import mkdtemp
import pysam

from dark.process import Executor
from dark.bowtie2 import Bowtie2

DEFAULT_SAMTOOLS_VIEW_FLAGS = (
    pysam.FUNMAP | pysam.FSECONDARY | pysam.FDUP | pysam.FSUPPLEMENTARY)


def saveStdin(args):
    if os.isatty(0):
        print('Reading sequences to match against from stdin.',
              file=sys.stderr)
    dirname = mkdtemp(prefix='run-bt2-stdin-', dir=args.tempdir)
    filename = join(dirname, 'stdin.fastq')
    count = 0

    with open(filename, 'w') as fp:
        for line in sys.stdin:
            fp.write(line)
            count += 1

    if count == 0:
        print('Standard input was empty! Exiting.', file=sys.stderr)
        sys.exit(1)

    return dirname, filename


def processMatch(args, e):
    """
    Run Bowtie2 to find matches.
    """
    if not args.index and not args.reference:
        print('One of --index or --reference must be given.', file=sys.stderr)
        sys.exit(0)

    if args.markDuplicatesPicard or args.callHaplotypesGATK:
        if args.picardJar:
            picardJar = args.picardJar
        else:
            try:
                picardJar = os.environ['PICARD_JAR']
            except KeyError:
                print('If you use --markDuplicatesPicard or '
                      '--callHaplotypesGATK, you must give a Picard JAR file '
                      'with --picardJar or else set PICARD_JAR in your '
                      'environment.', file=sys.stderr)
                sys.exit(0)

    bt2 = Bowtie2(
        executor=e,
        threads=(multiprocessing.cpu_count() if args.threads is None
                 else args.threads),
        verboseFp=(sys.stderr if args.verbose else None),
        dryRun=args.dryRun, tempdir=args.tempdir)

    bt2.buildIndex(args.index or args.reference)

    if not args.align:
        if args.verbose:
            print('Bowtie2 alignment not done due to --noAlign.',
                  file=sys.stderr)
        print('Bowtie2 temporary directory %r.' % bt2.tempdir, file=sys.stderr)
        sys.exit(0)

    if args.fastq1:
        stdinDir, fastq1, fastq2 = None, args.fastq1, args.fastq2
    else:
        stdinDir, fastq1 = saveStdin(args)
        fastq2 = None

    if args.out:
        if exists(args.out) and not (args.force or args.dryRun):
            print('Will not overwrite pre-existing output file %r. '
                  'Use --force to make me.' % args.out, file=sys.stderr)
            sys.exit(1)

        if args.indexBAM:
            bai = args.out + '.bai'
            if exists(bai) and not (args.force or args.dryRun):
                print('Will not overwrite pre-existing output file %r. '
                      'Use --force to make me.' % bai, file=sys.stderr)
                sys.exit(1)
            needBAI = True
        else:
            needBAI = False

    if args.callHaplotypesGATK:
        if args.vcfFile:
            for filename in (args.vcfFile, args.vcfFile + '.tbi'):
                if exists(filename) and not args.force:
                    print('Will not overwrite pre-existing VCF file %r. '
                          'Use --force to make me.' % filename,
                          file=sys.stderr)
                    sys.exit(1)

    if args.callHaplotypesBcftools:
        if args.vcfFile:
            if exists(args.vcfFile) and not args.force:
                print('Will not overwrite pre-existing VCF file %r. '
                      'Use --force to make me.' % filename, file=sys.stderr)
                sys.exit(1)

    bt2.align(bowtie2Args=args.bowtie2Args, fastq1=fastq1, fastq2=fastq2)

    if args.bam:
        bt2.makeBAM(args.samtoolsViewArgs)

    if args.sort and not (
            args.markDuplicatesPicard or args.markDuplicatesGATK):
        bt2.sort()

    if args.markDuplicatesPicard:
        bt2.sort()
        bt2.picard(picardJar)

    if args.markDuplicatesGATK:
        bt2.sort(byName=True)
        bt2.markDuplicatesGATK()

    if args.removeDuplicates:
        bt2.removeDuplicates()

    if args.callHaplotypesGATK:
        bt2.indexBAM()
        bt2.callHaplotypesGATK(picardJar=picardJar, vcfFile=args.vcfFile,
                               referenceFasta=args.reference)

    if args.callHaplotypesBcftools:
        bt2.indexBAM()
        bt2.callHaplotypesBcftools(vcfFile=args.vcfFile,
                                   referenceFasta=args.reference)

    if args.bam and args.indexBAM:
        bt2.indexBAM()

    if args.out:
        e.execute("mv '%s' '%s'" % (bt2.outputFile(), args.out))
        if needBAI:
            e.execute("mv '%s.bai' '%s.bai'" % (bt2.outputFile(), args.out))
    else:
        if not args.dryRun:
            with open(bt2.outputFile(), 'rb') as fp:
                read, write = fp.read, sys.stdout.write
                while True:
                    data = read(2048)
                    if data:
                        write(data)
                    else:
                        break

    if stdinDir:
        # Remove the directory where we stashed standard input.
        e.execute("rm -r '%s'" % stdinDir)

    if args.clean:
        bt2.close()
    else:
        print('Bowtie2 temporary directory %r.' % bt2.tempdir, file=sys.stderr)


def processOneIgnore(args, index, count, tempdir, e):
    """
    Process one ignored index.
    """
    if args.verbose:
        print('Preparing to ignore reads matching index %r.' % index,
              file=sys.stderr)

    bt2 = Bowtie2(
        executor=e,
        threads=(multiprocessing.cpu_count() if args.threads is None
                 else args.threads),
        verboseFp=(sys.stderr if args.verbose else None),
        dryRun=args.dryRun, tempdir=tempdir)

    bt2.buildIndex(index)

    unAlignedFile = join(tempdir, 'unaligned-%d' % count)

    if args.fastq1 and args.fastq2:
        bt2.align(
            bowtie2Args='%s --un-conc-gz %s' % (args.bowtie2Args,
                                                unAlignedFile),
            fastq1=args.fastq1, fastq2=args.fastq2, discardSAM=True)

        for i in '1', '2':
            src = unAlignedFile + '.' + i
            dst = src + '.gz'
            e.execute("mv '%s' '%s'" % (src, dst))
            setattr(args, 'fastq' + i, dst)

        if not e.dryRun:
            assert exists(args.fastq1)
            assert exists(args.fastq2)
    else:
        bt2.align(
            bowtie2Args='%s --un-gz %s' % (args.bowtie2Args,
                                           unAlignedFile),
            fastq1=args.fastq1, discardSAM=True)

        e.execute("mv '%s' '%s.gz'" % (unAlignedFile, unAlignedFile))
        args.fastq1 = unAlignedFile + '.gz'
        if not e.dryRun:
            assert exists(args.fastq1)

    if args.clean:
        bt2.close()
    else:
        print('Bowtie2 temporary ignore index directory %r.' % bt2.tempdir,
              file=sys.stderr)


def processIgnores(args, e):
    """
    Ignore the indices in args.ignoredIndices
    """
    tempdir = '/tmp/ignores' if e.dryRun else mkdtemp(
        prefix='bt2-ignores-', dir=args.tempdir)
    for count, ignoreIndex in enumerate(args.ignoredIndices):
        processOneIgnore(args, ignoreIndex, count, tempdir, e)
    return tempdir


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=('Run bowtie2 on a FASTA file. Optionally convert the '
                     'result to BAM, sorting, and indexing.'))

    parser.add_argument(
        '--index',
        help=('Either: an accession number, a filename or the name of a '
              'pre-existing bowtie2 index (created with bowtie2-build). If '
              'not given and --reference is used, the reference will be '
              'used to build a bowtie2 index.'))

    parser.add_argument(
        '--ignoreIndex', action='append', dest='ignoredIndices',
        help=('Either: an accession number, a filename or the name of a '
              'pre-existing bowtie2 index (created with bowtie2-build). '
              'Reads matching this index will be ignored. May be repeated.'))

    parser.add_argument(
        '--fastq1', '-1',
        help=('The FASTQ reads to match against the bowtie2 index given by '
              '--index. Also use --fast2 if you have paired reads. '
              'If not given, single-end FASTQ reads will be read from '
              'standard input.'))

    parser.add_argument(
        '--fastq2', '-2',
        help=('The FASTQ reads to match against the bowtie2 index given by '
              '--index. Use this with --fastq1 to specify the mate '
              'file for paired-end reads.'))

    parser.add_argument(
        '--bowtie2Args', default='--no-unal',
        help=('Extra arguments to be passed to Bowtie2 (use --threads to '
              'specify a thread count).'))

    parser.add_argument(
        '--samtoolsViewArgs',
        default='-F %d -q 30' % DEFAULT_SAMTOOLS_VIEW_FLAGS,
        help='Arguments to be passed to samtools view to create the BAM file.')

    parser.add_argument(
        '--tempdir',
        help=('The temporary directory to use. If not specified, the value '
              'of the TMPDIR environment variable (if any) is used, or else '
              '/tmp.'))

    parser.add_argument(
        '--out', '-o',
        help=('The output file name. If not given, the resulting SAM or BAM '
              'will be written to standard output will be used.'))

    parser.add_argument(
        '--reference',
        help=('The reference FASTA file for use with --callHaplotypesGATK and '
              '--callHaplotypesBcftools. This will be used to build a Bowtie2 '
              'index if --index is not given.'))

    parser.add_argument(
        '--vcfFile',
        help=('The file to write VCF info to if --callHaplotypesGATK or '
              '--callHaplotypesBcftools are used.'))

    parser.add_argument(
        '--markDuplicatesGATK', default=False, action='store_true',
        help=('Use GATK to mark duplicates. See '
              'https://gatk.broadinstitute.org for details on GATK.'))

    parser.add_argument(
        '--markDuplicatesPicard', default=False, action='store_true',
        help=('Use Picard to mark duplicates. See '
              'https://github.com/broadinstitute/picard for details on '
              'Picard.'))

    parser.add_argument(
        '--picardJar',
        help=('The path to the Picard jar file. See '
              'https://github.com/broadinstitute/picard for details on '
              'Picard.'))

    parser.add_argument(
        '--removeDuplicates', default=False, action='store_true',
        help=('Remove duplicates from the resulting SAM/BAM file. Best used '
              'in combination with an option that marks duplicates, such as '
              '--markDuplicatesGATK.'))

    parser.add_argument(
        '--verbose', default=False, action='store_true',
        help=('Print a description of commands as they are (or would be, if '
              '--dryRun is used) executed.'))

    parser.add_argument(
        '--log', default=False, action='store_true',
        help=('Show a log of commands that were (or would be, if --dryRun is '
              'used) executed.'))

    parser.add_argument(
        '--threads', type=int,
        help='The number of threads to use when running bowtie2 commands.')

    parser.add_argument(
        '--noAlign', default=True, action='store_false', dest='align',
        help='Do not align with Bowtie2, just build an index.')

    parser.add_argument(
        '--noBAM', default=True, action='store_false', dest='bam',
        help='Do not convert SAM to BAM.')

    parser.add_argument(
        '--noSort', default=True, action='store_false', dest='sort',
        help='Do not sort the BAM.')

    parser.add_argument(
        '--noIndexBAM', default=True, action='store_false', dest='indexBAM',
        help='Do not index the BAM file.')

    parser.add_argument(
        '--noClean', default=True, action='store_false', dest='clean',
        help='Do not remove intermediate files or the temporary directory.')

    parser.add_argument(
        '--force', default=False, action='store_true',
        help='Overwrite pre-existing output file.')

    parser.add_argument(
        '--dryRun', default=False, action='store_true',
        help='Do not run commands, just print what would be done.')

    haplotypeCaller = parser.add_mutually_exclusive_group()

    haplotypeCaller.add_argument(
        '--callHaplotypesGATK', default=False, action='store_true',
        help=('Use GATK to call haplotypes. See '
              'https://gatk.broadinstitute.org for details on GATK.'))

    haplotypeCaller.add_argument(
        '--callHaplotypesBcftools', default=False, action='store_true',
        help='Use bcftools call to call haplotypes.')

    args = parser.parse_args()

    if args.indexBAM and not args.bam:
        print('The --indexBAM option only makes sense if you do not use '
              '--noBAM.', file=sys.stderr)
        sys.exit(1)

    e = Executor(args.dryRun)

    if args.tempdir is None:
        args.tempdir = os.environ.get('TMPDIR', '/tmp')

    if args.ignoredIndices:
        ignoresDir = processIgnores(args, e)

    processMatch(args, e)

    if args.ignoredIndices:
        if args.clean:
            e.execute("rm -r '%s'" % ignoresDir)
        else:
            print('Temporary directory with non-ignored inputs %r.' %
                  ignoresDir, file=sys.stderr)

    if args.dryRun or args.log:
        print('\n'.join(e.log), file=sys.stderr)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

from __future__ import print_function, division

import argparse
from collections import defaultdict

from dark.sam import samfile

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Print SAM/BAM file reference sequence and unmapped read '
                 'counts.'))

parser.add_argument(
    'samFile', metavar='FILENAME',
    help='The name of a SAM/BAM alignment file.')

args = parser.parse_args()


def referenceInfo():
    return {
        'duplicateCount': 0,
        'primaryCount': 0,
        'qcFailCount': 0,
        'nonDuplicateCount': 0,
        'readIds': set(),
        'secondaryCount': 0,
        'supplementaryCount': 0,
    }


referenceReads = defaultdict(referenceInfo)
mappedCount = unmappedCount = 0
readIds = set()

with samfile(args.samFile) as fp:
    for read in fp.fetch():
        readIds.add(read.query_name)
        if read.is_unmapped:
            unmappedCount += 1
        else:
            mappedCount += 1
            stats = referenceReads[read.reference_name]
            stats['readIds'].add(read.query_name)
            if read.is_secondary:
                stats['secondaryCount'] += 1
            elif read.is_supplementary:
                stats['supplementaryCount'] += 1
            else:
                stats['primaryCount'] += 1
            if read.is_duplicate:
                stats['duplicateCount'] += 1
            else:
                stats['nonDuplicateCount'] += 1
            if read.is_qcfail:
                stats['qcFailCount'] += 1

totalReads = len(readIds)

print('Found a total of %d read%s (%d mapped, %d unmapped).' %
      (totalReads, '' if totalReads == 1 else 's', mappedCount, unmappedCount))


def pct(a, b):
    return (a / b if b else 0.0) * 100.0


for referenceId in sorted(referenceReads):
    stats = referenceReads[referenceId]
    readCount = len(stats['readIds'])
    print('%s: %d/%d (%.2f%%) reads mapped to the reference.\n'
          '  Non-duplicates: %d (%.2f%%), '
          'Duplicates: %d (%.2f%%), '
          'QC fails: %d (%.2f%%)\n'
          '  Primary: %d (%.2f%%), '
          'Secondary: %d (%.2f%%), '
          'Supplementary: %d (%.2f%%)' %
          (referenceId, readCount, totalReads, pct(readCount, totalReads),
           stats['nonDuplicateCount'],
           pct(stats['nonDuplicateCount'], readCount),
           stats['duplicateCount'],
           pct(stats['duplicateCount'], readCount),
           stats['qcFailCount'],
           pct(stats['qcFailCount'], readCount),
           stats['primaryCount'],
           pct(stats['primaryCount'], readCount),
           stats['secondaryCount'],
           pct(stats['secondaryCount'], readCount),
           stats['supplementaryCount'],
           pct(stats['supplementaryCount'], readCount)))



#!/usr/bin/env python

from __future__ import print_function, division

import sys
import warnings
from time import time
from itertools import chain
import argparse

from dark.civ.proteins import SqliteIndexWriter
from dark.taxonomy import (
    addTaxonomyDatabaseCommandLineOptions,
    parseTaxonomyDatabaseCommandLineOptions)


def filenamesAndAdders(args, db):
    """
    Get the filenames and database adding functions.

    @param args: A C{Namespace} instance as returned by argparse.
    @param db: An C{SqliteIndexWriter} instance.
    @return: A generator yielding 3-tuples containing a filename, a
        database adder function that can read the file and incorporate it,
        and a C{str} that is either 'gb' or 'json'.
    """
    # Use chain to flatten the lists of lists that we get from using both
    # nargs='+' and action='append'. We use both because it allows people
    # to use (e.g.) --gb on the command line either via "--gb file1 --gb
    # file2" or "--gb file1 file2", or a combination of these. That way
    # it's not necessary to remember which way you're supposed to use it
    # and you also can't be hit by the subtle problem encountered in
    # https://github.com/acorg/dark-matter/issues/453

    if not (args.gb or args.json):
        print('At least one of --gb or --json must be used to indicate input '
              'files to process.', file=sys.stderr)
        sys.exit(1)

    if args.gb:
        for filename in chain.from_iterable(args.gb):
            yield filename, db.addGenBankFile, 'gb'

    if args.json:
        for filename in chain.from_iterable(args.json):
            yield filename, db.addJSONFile, 'json'


def main(args, parser):
    """
    Build the protein database.

    @param args: The namespace of command-line arguments returned by
        argparse.parse_args()
    @param parser: An C{argparse.ArgumentParser} instance.
    """

    if (args.minGenomeLength is not None and
            args.maxGenomeLength is not None and
            args.minGenomeLength > args.maxGenomeLength):
        raise ValueError(
            '--minGenomeLength cannot be larger than --maxGenomeLength')

    if args.excludeExclusiveHost:
        excludeExclusiveHosts = set(chain.from_iterable(
            args.excludeExclusiveHost))
    else:
        excludeExclusiveHosts = None

    taxonomyDatabase = parseTaxonomyDatabaseCommandLineOptions(args, parser)
    progress = args.progress

    if progress:
        overallStart = time()
        totalGenomeCount = totalProteinCount = 0

    with SqliteIndexWriter(args.databaseFile) as db:
        for fileCount, (filename, addFunc, type_) in enumerate(
                filenamesAndAdders(args, db), start=1):

            if args.logFile:
                print("\n>>> Indexing '%s'." % filename, end='\n\n',
                      file=args.logFile)

            if progress:
                start = time()

            examinedGenomeCount, genomeCount, proteinCount = addFunc(
                filename, dnaOnly=args.dnaOnly, rnaOnly=args.rnaOnly,
                minGenomeLength=args.minGenomeLength,
                maxGenomeLength=args.maxGenomeLength,
                excludeExclusiveHosts=excludeExclusiveHosts,
                excludeFungusOnlyViruses=args.excludeFungusOnlyViruses,
                excludePlantOnlyViruses=args.excludePlantOnlyViruses,
                databaseName=args.databaseName,
                taxonomyDatabase=taxonomyDatabase,
                proteinSource=args.proteinSource,
                genomeSource=args.genomeSource,
                duplicationPolicy=args.duplicationPolicy, logfp=args.logFile)

            if examinedGenomeCount == 0:
                if type_ == 'gb':
                    print('WARNING: No genomes found in %r. Did the GenBank '
                          'download fail on that file?' % filename,
                          file=sys.stderr)
                else:
                    assert type_ == 'json'
                    print('WARNING: no genomes found in JSON file %r.' %
                          filename, file=sys.stderr)

            if progress:
                elapsed = time() - start
                totalGenomeCount += genomeCount
                totalProteinCount += proteinCount
                print('Processed %r: added %3d of %3d genome%s (%5d '
                      'protein%s) in %.2f seconds.' %
                      (filename, genomeCount, examinedGenomeCount,
                       ' ' if examinedGenomeCount == 1 else 's',
                       proteinCount, '' if proteinCount == 1 else 's',
                       elapsed), file=sys.stderr)

    if progress:
        elapsed = time() - overallStart
        print('%d files (containing %d genomes and %d proteins) '
              'indexed in %.2f seconds (%.2f mins).' %
              (fileCount, totalGenomeCount, totalProteinCount, elapsed,
               elapsed / 60), file=sys.stderr)


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Create a genome/protein sqlite3 database from GenBank '
                 'and JSON files. The protein sequences for the database are '
                 'printed to standard output for indexing by other tools.'))

parser.add_argument(
    '--databaseFile', required=True,
    help=('The output file. This file must not exist (use --force to '
          'overwrite).'))

parser.add_argument(
    '--databaseName',
    help=('The database that the records in the (--gb) GenBank files came '
          'from (e.g., "refseq" or "RVDB").'))

parser.add_argument(
    '--duplicationPolicy', choices=('error', 'ignore'), default='ignore',
    help=('What to do if a to-be-inserted accession number is already '
          'present in the database. "error" results in a ValueError being '
          'raised, "ignore" will ignore the duplicate. It should also be '
          'possible to update (i.e., replace) but that is not supported yet.'))

parser.add_argument(
    '--progress', default=False, action='store_true',
    help='Print indexing progress.')

parser.add_argument(
    '--logFile', type=argparse.FileType('w'),
    help='Write indexing details to a log file.')

parser.add_argument(
    '--noWarnings', default=False, action='store_true',
    help='Do not print warnings about unparseable GenBank or JSON records.')

# A mutually exclusive group for DNA/RNA only.
group = parser.add_mutually_exclusive_group()

group.add_argument(
    '--dnaOnly', default=False, action='store_true',
    help='If given, only include DNA viruses.')

group.add_argument(
    '--rnaOnly', default=False, action='store_true',
    help='If given, only include RNA viruses.')

parser.add_argument(
    '--maxGenomeLength', type=int,
    help='Genomes longer than this will not be considered.')

parser.add_argument(
    '--minGenomeLength', type=int,
    help='Genomes shorter than this will not be considered.')

parser.add_argument(
    '--excludeFungusOnlyViruses', default=False, action='store_true',
    help=('If given, exclude fungus-only viruses (i.e., viruses that only '
          'infect fungi host species).'))

parser.add_argument(
    '--excludePlantOnlyViruses', default=False, action='store_true',
    help=('If given, exclude plant-only viruses (i.e., viruses that only '
          'infect plant host species).'))

parser.add_argument(
    '--gb', metavar='GenBank-file', nargs='+', action='append',
    help=('The GenBank file(s) to make the database from. These may be '
          'uncompressed, or compressed with bgzip (from samtools), with '
          'a .gz suffix.'))

parser.add_argument(
    '--json', metavar='JSON-file', nargs='+', action='append',
    help=('The JSON file(s) to make the database from. These contain '
          'genome and protein information for cases where sequences that '
          'are not in GenBank should be added.'))

parser.add_argument(
    '--proteinSource', default='GENBANK',
    help=('The source of the accession numbers for the proteins found in the '
          'input files. This becomes part of the sequence id printed in the '
          'protein FASTA output.'))

parser.add_argument(
    '--genomeSource', default='GENBANK',
    help=('The source of the accession numbers for the genomes in the input '
          'files. This becomes part of the sequence id printed in the '
          'protein FASTA output.'))

parser.add_argument(
    '--excludeExclusiveHost', metavar='hostname', nargs='+', action='append',
    choices=('algae', 'archaea', 'bacteria', 'diatom', 'environment',
             'eukaryotes', 'fungi', 'human', 'invertebrates', 'plants',
             'protozoa', 'vertebrates'),
    help=('A host type that should be excluded, but only if this is the only '
          'host of the virus.'))

addTaxonomyDatabaseCommandLineOptions(parser)

args = parser.parse_args()

if args.noWarnings:
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        main(args)
else:
    main(args, parser)



# From https://samtools.github.io/hts-specs/SAMv1.pdf
CINS, CDEL, CMATCH, CEQUAL, CDIFF = 'IDM=X'


def dna2cigar(s1, s2, concise=False):
    """
    Form a CIGAR string from two equal-length DNA sequences. Case is not
    ignored, so if you want to ignore case you must convert the sequences
    to a consistent case (upper or lower, doesn't matter) before calling.

    @param s1: A C{str} of DNA 'A', 'G', 'C', 'T' characters.
    @param s2: A C{str} of DNA 'A', 'G', 'C', 'T' characters.
    @param concise: If C{True}, use 'M' for matches and mismatches instead
        of the more specific 'X' and '='.
    @raise ValueError: If the two sequences are not of equal length or if the
        sequences are of length zero.
    @return: A C{str} CIGAR string.
    """
    len1 = len(s1)
    if len1 != len(s2):
        raise ValueError('Sequences %r and %r of unequal length (%d != %d).' %
                         (s1, s2, len1, len(s2)))

    if len1 == 0:
        raise ValueError('Two sequences of zero length were passed.')

    if concise:
        return '%d%s' % (len1, CMATCH)

    result = []
    length, operation = 0, None

    for n1, n2 in zip(s1, s2):
        if n1 == n2:
            if operation == CEQUAL:
                length += 1
            else:
                if length:
                    assert operation == CDIFF
                    result.append('%d%s' % (length, CDIFF))
                length, operation = 1, CEQUAL
        else:
            if operation == CDIFF:
                length += 1
            else:
                if length:
                    assert operation == CEQUAL
                    result.append('%d%s' % (length, CEQUAL))
                length, operation = 1, CDIFF

    # Append the final operation.
    result.append('%d%s' % (length, operation))

    return ''.join(result)



from unittest import TestCase
from io import StringIO

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from dark.bowtie2 import Bowtie2
from dark.process import Executor


class TestBowtie2(TestCase):
    """
    Test the Bowtie2 class.
    """
    def testIndexAccession(self):
        """
        Making a Bowtie index from an accession number must result in the
        expected commands being run and output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('MN908947.3')
        self.assertEqual(
            "$ bowtie2-build --quiet '/tmp/xxx/MN908947.3.fasta' "
            "'/tmp/xxx/index'",
            e.log[-1])
        log = fp.getvalue()
        self.assertTrue(
            log.startswith('Downloading FASTA for accession MN908947.3 '
                           'from NCBI.\n'))

    @patch('os.path.exists')
    def testIndexFromFile(self, existsMock):
        """
        Making a Bowtie index from a file name must result in the
        expected commands being run and output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        self.assertEqual(
            "$ bowtie2-build --quiet 'file.fasta' '/tmp/xxx/index'",
            e.log[-1])
        log = fp.getvalue()
        self.assertTrue(
            log.find("Building Bowtie2 index from file.fasta.\n") > -1)

    @patch('os.path.exists')
    def testIndexFromBowtie2File(self, existsMock):
        """
        Making a Bowtie index from a file that is an existing Bowtie2 index
        file must result in the expected commands being run and output being
        produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.1.bt2')
        self.assertEqual(-1, e.log[-1].find('bowtie2-build'))
        log = fp.getvalue()
        self.assertEqual("Using pre-existing Bowtie2 index 'file'.\n", log)

    @patch('os.path.exists')
    def testIndexFromBowtie2FilePrefix(self, existsMock):
        """
        Making a Bowtie index from a file that is a prefix of an existing
        Bowtie2 index file name must result in the expected commands being run
        and output being produced.
        """
        class SideEffect(object):
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename):
                if self.count == 0:
                    self.test.assertEqual('idx-file', filename)
                    self.count += 1
                    return False
                elif self.count == 1:
                    self.test.assertEqual('idx-file.1.bt2', filename)
                    self.count += 1
                    return True
                else:
                    self.test.fail('Unexpected third call to exists.')

        existsMock.side_effect = SideEffect(self).sideEffect

        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('idx-file')
        self.assertEqual(-1, e.log[-1].find('bowtie2-build'))
        log = fp.getvalue()
        self.assertEqual("Using pre-existing Bowtie2 index 'idx-file'.\n", log)

    @patch('os.path.exists')
    def testAlignOneFASTQ(self, existsMock):
        """
        Making a Bowtie index from a file name and running an alignment with
        one FASTQ file must result in the expected commands being run and
        output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq', threads=4)
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nAligning with Bowtie2.\n'))
        self.assertEqual(
            "$ bowtie2 --no-unal --threads 4 --rg-id 'orig' --rg 'SM:orig' "
            "-x '/tmp/xxx/index' "
            "-U 'file1.fastq' > '/tmp/xxx/result.sam'", e.log[-1])

    @patch('os.path.exists')
    def testAlignOneFASTQWithReadGroup(self, existsMock):
        """
        Making a Bowtie index from a file name and running an alignment with
        one FASTQ file and specifying a read group must result in the expected
        commands being run and output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq', threads=4, readGroup='xxx')
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nAligning with Bowtie2.\n'))
        self.assertEqual(
            "$ bowtie2 --no-unal --threads 4 --rg-id 'xxx' --rg 'SM:orig' "
            "-x '/tmp/xxx/index' "
            "-U 'file1.fastq' > '/tmp/xxx/result.sam'", e.log[-1])

    @patch('os.path.exists')
    def testAlignOneFASTQWithSampleName(self, existsMock):
        """
        Making a Bowtie index from a file name and running an alignment with
        one FASTQ file and specifying a sample name must result in the expected
        commands being run and output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq', threads=4, sampleName='xxx')
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nAligning with Bowtie2.\n'))
        self.assertEqual(
            "$ bowtie2 --no-unal --threads 4 --rg-id 'orig' --rg 'SM:xxx' "
            "-x '/tmp/xxx/index' "
            "-U 'file1.fastq' > '/tmp/xxx/result.sam'", e.log[-1])

    @patch('os.path.exists')
    def testAlignOneFASTQWithFlags(self, existsMock):
        """
        Making a Bowtie index from a file name and running an alignment with
        one FASTQ file and some (fictional) args must result in the expected
        commands being run and output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq', threads=4, bowtie2Args='--up --dn')
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nAligning with Bowtie2.\n'))
        self.assertEqual(
            "$ bowtie2 --up --dn --threads 4 --rg-id 'orig' --rg 'SM:orig' -x "
            "'/tmp/xxx/index' -U 'file1.fastq' > '/tmp/xxx/result.sam'",
            e.log[-1])

    @patch('os.path.exists')
    def testAlignTwoFASTQs(self, existsMock):
        """
        Making a Bowtie index from a file name and running an alignment with
        two FASTQ files must result in the expected commands being run and
        output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq', fastq2='file2.fastq', threads=4)
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nAligning with Bowtie2.\n'))
        self.assertEqual(
            "$ bowtie2 --no-unal --threads 4 --rg-id 'orig' --rg 'SM:orig' -x "
            "'/tmp/xxx/index' "
            "-1 'file1.fastq' -2 'file2.fastq' > '/tmp/xxx/result.sam'",
            e.log[-1])

    @patch('os.path.exists')
    def testAlignTwoFASTQsWithReadGroup(self, existsMock):
        """
        Making a Bowtie index from a file name and running an alignment with
        two FASTQ files and a read group must result in the expected commands
        being run and output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq', fastq2='file2.fastq', threads=4,
                 readGroup='xxx')
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nAligning with Bowtie2.\n'))
        self.assertEqual(
            "$ bowtie2 --no-unal --threads 4 --rg-id 'xxx' --rg 'SM:orig' -x "
            "'/tmp/xxx/index' "
            "-1 'file1.fastq' -2 'file2.fastq' > '/tmp/xxx/result.sam'",
            e.log[-1])

    @patch('os.path.exists')
    def testAlignTwoFASTQsWithSampleName(self, existsMock):
        """
        Making a Bowtie index from a file name and running an alignment with
        two FASTQ files and a sample name must result in the expected commands
        being run and output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq', fastq2='file2.fastq', threads=4,
                 sampleName='xxx')
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nAligning with Bowtie2.\n'))
        self.assertEqual(
            "$ bowtie2 --no-unal --threads 4 --rg-id 'orig' --rg 'SM:xxx' -x "
            "'/tmp/xxx/index' "
            "-1 'file1.fastq' -2 'file2.fastq' > '/tmp/xxx/result.sam'",
            e.log[-1])

    @patch('os.path.exists')
    def testMakeBAM(self, existsMock):
        """
        Making a BAM file must result in the expected commands being run and
        output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq')
        bt.makeBAM()
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nConverting SAM to BAM.\n'))
        self.assertEqual(
            "$ samtools view -b  '/tmp/xxx/result.sam' > "
            "'/tmp/xxx/result.bam'",
            e.log[-1])

    @patch('os.path.exists')
    def testMakeIndexedBAM(self, existsMock):
        """
        Making a BAM file and indexing it must result in the expected commands
        being run and output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq')
        bt.makeBAM()
        bt.indexBAM()
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nIndexing BAM.\n'))
        self.assertEqual("$ samtools index '/tmp/xxx/result.bam'", e.log[-1])

    @patch('os.path.exists')
    def testRemoveDuplicates(self, existsMock):
        """
        Removing duplicates must result in the expected commands
        being run and output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq')
        bt.removeDuplicates()
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nRemoving marked duplicates.\n'))
        self.assertEqual("$ samtools view -b -F 1024 '/tmp/xxx/result.sam' "
                         "> '/tmp/xxx/non-duplicates.sam'", e.log[-2])
        self.assertEqual("$ mv '/tmp/xxx/non-duplicates.sam' "
                         "'/tmp/xxx/result.sam'", e.log[-1])

    @patch('os.path.exists')
    def testSortByCoord(self, existsMock):
        """
        Sorting by coord (the default) must result in the expected commands
        being run and output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq')
        bt.sort()
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nSorting SAM (by coord).\n'))
        self.assertEqual("$ samtools sort '/tmp/xxx/result.sam' > "
                         "'/tmp/xxx/result-sorted.sam'", e.log[-2])
        self.assertEqual("$ mv '/tmp/xxx/result-sorted.sam' "
                         "'/tmp/xxx/result.sam'", e.log[-1])

    @patch('os.path.exists')
    def testSortByName(self, existsMock):
        """
        Sorting by name must result in the expected commands
        being run and output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq')
        bt.sort(byName=True)
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nSorting SAM (by name).\n'))
        self.assertEqual("$ samtools sort -n '/tmp/xxx/result.sam' > "
                         "'/tmp/xxx/result-sorted.sam'", e.log[-2])
        self.assertEqual("$ mv '/tmp/xxx/result-sorted.sam' "
                         "'/tmp/xxx/result.sam'", e.log[-1])

    @patch('os.path.exists')
    def testOutputFileSAM(self, existsMock):
        """
        The output file must be a SAM file if BAM hasn't been produced.
        """
        class SideEffect(object):
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename):
                if self.count == 0:
                    self.test.assertEqual('index', filename)
                    self.count += 1
                    return False
                elif self.count == 1:
                    self.test.assertEqual('index.1.bt2', filename)
                    self.count += 1
                    return True
                elif self.count == 2:
                    self.test.assertEqual('/tmp/xxx/result.bam', filename)
                    self.count += 1
                    return False
                elif self.count == 3:
                    self.test.assertEqual('/tmp/xxx/result.sam', filename)
                    self.count += 1
                    return True
                else:
                    self.test.fail(
                        'Unexpected 5th call to exists. Filename: %r.' %
                        filename)

        existsMock.side_effect = SideEffect(self).sideEffect

        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('index')
        bt.align(fastq1='file1.fastq')
        self.assertEqual('/tmp/xxx/result.sam', bt.outputFile())

    @patch('os.path.exists')
    def testOutputFileBAM(self, existsMock):
        """
        The output file must be BAM if it has been produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq')
        bt.makeBAM()
        self.assertEqual('/tmp/xxx/result.bam', bt.outputFile())

    @patch('os.path.exists')
    def testMarkDuplicatesPicard(self, existsMock):
        """
        Using Picard to mark duplicates must result in the expected commands
        being run and output being produced.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq')
        bt.makeBAM()
        bt.indexBAM()
        bt.markDuplicatesPicard('picard.jar')
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nMarking duplicates with Picard.\n'))
        self.assertEqual(
            '$ java -Xmn2g -Xms2g -Xmx2g -jar picard.jar MarkDuplicates '
            "I='/tmp/xxx/result.bam' O='/tmp/xxx/picard-duplicates.bam' "
            "M=/dev/null >'/tmp/xxx/picard.errs' 2>&1",
            e.log[-2])
        self.assertEqual("$ mv '/tmp/xxx/picard-duplicates.bam' "
                         "'/tmp/xxx/result.bam'", e.log[-1])

    @patch('os.path.exists')
    def testMarkDuplicatesGATKBowtie2Threads(self, existsMock):
        """
        Using GATK to mark duplicates must result in the expected commands
        being run and output being produced. The number of threads is taken
        from the Bowtie2 instance.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp, threads=4)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq')
        bt.makeBAM()
        bt.indexBAM()
        bt.markDuplicatesGATK()
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nMarking duplicates with GATK.\n'))
        self.assertEqual(
            "$ gatk MarkDuplicatesSpark -I '/tmp/xxx/result.bam' -O "
            "'/tmp/xxx/gatk-duplicates.bam' --conf spark.executor.cores=4",
            e.log[-2])
        self.assertEqual("$ mv '/tmp/xxx/gatk-duplicates.bam' "
                         "'/tmp/xxx/result.bam'", e.log[-1])

    @patch('os.path.exists')
    def testMarkDuplicatesGATKThreadsInArg(self, existsMock):
        """
        Using GATK to mark duplicates must result in the expected commands
        being run and output being produced. The number of threads is given
        in the call to markDuplicatesGATK and must override the number given
        to the Bowtie2 instance.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp, threads=4)
        bt.buildIndex('file.fasta')
        bt.align(fastq1='file1.fastq')
        bt.makeBAM()
        bt.indexBAM()
        bt.markDuplicatesGATK(threads=32)
        log = fp.getvalue()
        self.assertTrue(log.endswith('\nMarking duplicates with GATK.\n'))
        self.assertEqual(
            "$ gatk MarkDuplicatesSpark -I '/tmp/xxx/result.bam' -O "
            "'/tmp/xxx/gatk-duplicates.bam' --conf spark.executor.cores=32",
            e.log[-2])
        self.assertEqual("$ mv '/tmp/xxx/gatk-duplicates.bam' "
                         "'/tmp/xxx/result.bam'", e.log[-1])

    def testClose(self):
        """
        Calling close() must result in the temporary directory being removed.
        """
        e = Executor(dryRun=True)
        fp = StringIO()
        bt = Bowtie2(executor=e, dryRun=True, verboseFp=fp)
        bt.close()
        self.assertEqual("$ rm -r '/tmp/xxx'", e.log[-1])



from six import PY3
from Bio import SeqIO

from dark.reads import Reads, SSAARead
from dark.utils import asHandle


class SSFastaReads(Reads):
    """
    Subclass of L{dark.reads.Reads} providing access to reads in FASTA format
    along with their PDB secondary structure.

    The pdb file is in FASTA format. For each pair of sequence / structure,
    it gives the amino acid sequence in the first record and the predicted
    secondary structure in the next. This is the file format used to hold
    results from the DSSP program. Secondary structures found by DSSP can be
    downloaded from http://www.rcsb.org/pdb/files/ss.txt on 11/11/2015

    IMPORTANT NOTE: the ss.txt file contains spaces in the structure
    records.  SeqIO.parse will silently collapse these to nothing, which
    will result in unequal length sequence and structure strings. So you
    will need to replace the spaces in that file with something else, like
    '-', to make sure the structure information has the correct length and
    alignment with the sequence.

    @param _files: Either a single C{str} file name or file handle, or a
        C{list} of C{str} file names and/or file handles. Each file or file
        handle must contain sequences in PDB FASTA format (see above).
    @param readClass: The class of read that should be yielded by iter. This
        must accept 3 C{str} arguments: an id, the sequence, the structure.
    @param upperCase: If C{True}, both read and structure sequences will be
        converted to upper case.
    """
    def __init__(self, _files, readClass=SSAARead, upperCase=False):
        self._files = _files if isinstance(_files, (list, tuple)) else [_files]
        self._readClass = readClass
        self._upperCase = upperCase
        if PY3:
            super().__init__()
        else:
            Reads.__init__(self)

    def iter(self):
        """
        Iterate over the sequences in self.file_, yielding each as an
        instance of the desired read class.

        @raise ValueError: If the input file has an odd number of records or
            if any sequence has a different length than its predicted
            secondary structure.
        """
        upperCase = self._upperCase
        for _file in self._files:
            with asHandle(_file) as fp:
                records = SeqIO.parse(fp, 'fasta')
                while True:
                    try:
                        record = next(records)
                    except StopIteration:
                        break
                    try:
                        structureRecord = next(records)
                    except StopIteration:
                        raise ValueError('Structure file %r has an odd number '
                                         'of records.' % _file)

                    if len(structureRecord) != len(record):
                        raise ValueError(
                            'Sequence %r length (%d) is not equal to '
                            'structure %r length (%d) in input file %r.' % (
                                record.description, len(record),
                                structureRecord.description,
                                len(structureRecord), _file))

                    if upperCase:
                        read = self._readClass(
                            record.description,
                            str(record.seq.upper()),
                            str(structureRecord.seq.upper()))
                    else:
                        read = self._readClass(record.description,
                                               str(record.seq),
                                               str(structureRecord.seq))

                    yield read



from unittest import TestCase

from dark.score import HigherIsBetterScore, LowerIsBetterScore


class TestHigherIsBetterScore(TestCase):
    """
    Tests of the C{dark.score.HigherIsBetterScore} class.
    """

    def testExpectedAttr(self):
        """
        The expected score attribute must be present.
        """
        score = HigherIsBetterScore(10)
        self.assertEqual(10, score.score)

    def testEqual(self):
        """
        __eq__ must work as expected.
        """
        self.assertEqual(HigherIsBetterScore(10), HigherIsBetterScore(10))

    def testLessThan(self):
        """
        __lt__ must work as expected.
        """
        self.assertTrue(HigherIsBetterScore(10) < HigherIsBetterScore(20))

    def testBetterThan(self):
        """
        betterThan must work as expected.
        """
        self.assertTrue(HigherIsBetterScore(10).betterThan(5))


class TestLowerIsBetterScore(TestCase):
    """
    Tests of the C{dark.score.LowerIsBetterScore} class.
    """

    def testExpectedAttr(self):
        """
        The expected score attribute must be present.
        """
        score = LowerIsBetterScore(10)
        self.assertEqual(10, score.score)

    def testEqual(self):
        """
        __eq__ must work as expected.
        """
        self.assertEqual(LowerIsBetterScore(10), LowerIsBetterScore(10))

    def testLessThan(self):
        """
        __lt__ must work as expected.
        """
        self.assertTrue(LowerIsBetterScore(20) < LowerIsBetterScore(10))

    def testBetterThan(self):
        """
        betterThan must work as expected.
        """
        self.assertTrue(LowerIsBetterScore(10).betterThan(50))



import sys
from os.path import join
from subprocess import CalledProcessError
from tempfile import mkdtemp
from shutil import rmtree

from dark.fasta import FastaReads
from dark.process import Executor
from dark.reads import Reads


def mafft(reads, verbose=False, options=None, threads=None):
    """
    Run a MAFFT alignment and return the sequences.

    @param reads: An iterable of multiple reads.
    @param verbose: If C{True} print progress info to sys.stderr.
    @param options: A C{str} of options to pass to mafft.
    @return: A C{Reads} instance with the aligned sequences.
    """
    tempdir = mkdtemp()

    infile = join(tempdir, 'input.fasta')
    out = join(tempdir, 'result.fasta')

    Reads(reads).save(infile)

    if verbose:
        print('Running mafft.', file=sys.stderr)

    Executor().execute("mafft %s %s '%s' > '%s'" % (
        ('' if threads is None else '--thread %d' % threads),
        options or '', infile, out))

    # Use 'list' in the following to force reading the FASTA from disk.
    result = Reads(list(FastaReads(out)))
    rmtree(tempdir)

    return result


def needle(reads, verbose=False, options=None):
    """
    Run a Needleman-Wunsch alignment and return the two sequences.

    @param reads: An iterable of two reads.
    @param verbose: If C{True} print progress info to sys.stderr.
    @param options: Additional options to pass to needle.
    @return: A C{Reads} instance with the two aligned sequences.
    """
    tempdir = mkdtemp()

    file1 = join(tempdir, 'file1.fasta')
    with open(file1, 'w') as fp:
        print(reads[0].toString('fasta'), end='', file=fp)

    file2 = join(tempdir, 'file2.fasta')
    with open(file2, 'w') as fp:
        print(reads[1].toString('fasta'), end='', file=fp)

    out = join(tempdir, 'result.fasta')

    def useStderr(e):
        return "Sequences too big. Try 'stretcher'" not in e.stderr

    if verbose:
        print('Running needle.', file=sys.stderr)
    try:
        Executor().execute(
            "needle -asequence '%s' -bsequence '%s' %s "
            "-outfile '%s' -aformat fasta" % (
                file1, file2, options or '', out), useStderr=useStderr)
    except CalledProcessError as e:
        if useStderr(e):
            raise
        else:
            if verbose:
                print('Sequences too long for needle. Falling back to '
                      'stretcher. Be patient!', file=sys.stderr)
            Executor().execute("stretcher -asequence '%s' -bsequence '%s' "
                               "-auto "
                               "-outfile '%s' -aformat fasta" % (
                                   file1, file2, out))

    # Use 'list' in the following to force reading the FASTA from disk.
    result = Reads(list(FastaReads(out)))
    rmtree(tempdir)

    return result



#!/usr/bin/env python

from __future__ import print_function

import sys
import argparse
import re

from dark.taxonomy import (
    addTaxonomyDatabaseCommandLineOptions,
    parseTaxonomyDatabaseCommandLineOptions)

VERSION_REGEX = re.compile(r'\.\d+$')


def hosts(id_, db):
    """
    Look up hosts for an accession or taxonomy id.

    @param id_: The C{str} accession number or taxonomy id to look up.
    @param db: The sqlite3 taxonomy database to consult.
    @return: A C{set} of C{str} host names.
    """
    try:
        idInt = int(id_)
    except ValueError:
        hosts = db.hosts(id_)
    else:
        hosts = db.hosts(idInt)

    if hosts is None and VERSION_REGEX.search(id_) is None:
        # Try adding a version number.
        hosts = db.hosts(id_ + '.1')

    return hosts


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Print hosts for accession numbers or taxonomy ids')

    parser.add_argument(
        'ids', nargs='*',
        help=('The ids (accession numbers, names, or taxonomy ids) to print '
              'host information for. If not given, ids are read from '
              'standard input, one per line.'))

    parser.add_argument(
        '--database', required=True,
        help=('The file holding the sqlite3 taxonomy database. See '
              'https://github.com/acorg/ncbi-taxonomy-database for how to '
              'build one.'))

    parser.add_argument(
        '--printId', default=False, action='store_true',
        help='If specified, also print the id.')

    addTaxonomyDatabaseCommandLineOptions(parser)

    args = parser.parse_args()

    db = parseTaxonomyDatabaseCommandLineOptions(args, parser)

    if args.ids:
        ids = args.ids
    else:
        ids = (line[:-1] for line in sys.stdin)

    for id_ in ids:
        if args.printId:
            print(id_ + ':')
        hosts = hosts(id_, db)
        if hosts:
            print(', '.join(sorted(hosts)))
        else:
            print('No host information for %r found in the taxonomy database.'
                  % id_, file=sys.stderr)



from six.moves import builtins

from dark.reads import AARead, DNARead, RNARead
from dark.fastq import FastqReads
from dark.utils import StringIO

from unittest import TestCase

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from .mocking import mockOpen


class TestFastqReads(TestCase):
    """
    Tests for the L{dark.fastq.FastqReads} class.
    """

    def testEmpty(self):
        """
        An empty FASTQ file results in an empty iterator.
        """
        mockOpener = mockOpen()
        with patch.object(builtins, 'open', mockOpener):
            reads = FastqReads('filename.fastq')
            self.assertEqual([], list(reads))

    def testOneRead(self):
        """
        A FASTQ file with one read must be read properly.
        """
        data = '\n'.join(['@id1', 'ACGT', '+', '!!!!'])
        mockOpener = mockOpen(read_data=data)
        with patch.object(builtins, 'open', mockOpener):
            reads = list(FastqReads('filename.fastq'))
            self.assertEqual([DNARead('id1', 'ACGT', '!!!!')], reads)

    def testTwoReads(self):
        """
        A FASTQ file with two reads must be read properly and its
        sequences must be returned in the correct order.
        """
        data = '\n'.join(['@id1', 'ACGT', '+', '!!!!',
                          '@id2', 'TGCA', '+', '????'])
        mockOpener = mockOpen(read_data=data)
        with patch.object(builtins, 'open', mockOpener):
            reads = list(FastqReads('filename.fastq'))
            self.assertEqual(2, len(reads))
            self.assertEqual([DNARead('id1', 'ACGT', '!!!!'),
                              DNARead('id2', 'TGCA', '????')], reads)

    def testTypeDefaultsToDNA(self):
        """
        A FASTQ file whose type is not specified must result in reads that
        are instances of DNARead.
        """
        data = '\n'.join(['@id1', 'ACGT', '+', '!!!!'])
        mockOpener = mockOpen(read_data=data)
        with patch.object(builtins, 'open', mockOpener):
            reads = list(FastqReads('filename.fastq'))
            self.assertTrue(isinstance(reads[0], DNARead))

    def testTypeAA(self):
        """
        A FASTQ file whose read class is AARead must result in reads that
        are instances of AARead.
        """
        data = '\n'.join(['@id1', 'ACGT', '+', '!!!!'])
        mockOpener = mockOpen(read_data=data)
        with patch.object(builtins, 'open', mockOpener):
            reads = list(FastqReads('filename.fastq', AARead))
            self.assertTrue(isinstance(reads[0], AARead))

    def testTypeDNA(self):
        """
        A FASTQ file whose read class is DNARead must result in reads that
        are instances of DNARead.
        """
        data = '\n'.join(['@id1', 'ACGT', '+', '!!!!'])
        mockOpener = mockOpen(read_data=data)
        with patch.object(builtins, 'open', mockOpener):
            reads = list(FastqReads('filename.fastq', DNARead))
            self.assertTrue(isinstance(reads[0], DNARead))

    def testTypeRNA(self):
        """
        A FASTQ file whose read class is RNARead must result in reads that
        are instances of RNARead.
        """
        data = '\n'.join(['@id1', 'ACGT', '+', '!!!!'])
        mockOpener = mockOpen(read_data=data)
        with patch.object(builtins, 'open', mockOpener):
            reads = list(FastqReads('filename.fastq', RNARead))
            self.assertTrue(isinstance(reads[0], RNARead))

    def testTwoFiles(self):
        """
        It must be possible to read from two FASTQ files.
        """
        class SideEffect(object):
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename):
                if self.count == 0:
                    self.test.assertEqual('file1.fastq', filename)
                    self.count += 1
                    return StringIO('@id1\nACTG\n+\n!!!!\n')
                elif self.count == 1:
                    self.test.assertEqual('file2.fastq', filename)
                    self.count += 1
                    return StringIO('@id2\nCAGT\n+\n!!!!\n')
                else:
                    self.test.fail('We are only supposed to be called twice!')

        sideEffect = SideEffect(self)
        with patch.object(builtins, 'open') as mockMethod:
            mockMethod.side_effect = sideEffect.sideEffect
            reads = FastqReads(['file1.fastq', 'file2.fastq'])
            self.assertEqual(
                [
                    DNARead('id1', 'ACTG', '!!!!'),
                    DNARead('id2', 'CAGT', '!!!!'),
                ],
                list(reads))



#!/usr/bin/env python

from __future__ import print_function

import sys
import tempfile

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description=('Given FASTA sequences (of equal length) on stdin '
                     'write Phylip to stdout.'))

    parser.add_argument(
        '--addIdSpace', action='store_true', default=False,
        help=('If True, print an extra space after each sequence id (the '
              'space is expected by some programs that process Phylip files, '
              'e.g., baseml).'))

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    idSpace = ' ' if args.addIdSpace else ''
    reads = parseFASTACommandLineOptions(args)

    count = 0
    length = None
    with tempfile.TemporaryFile() as fp:
        for read in reads:
            count += 1
            if length is None:
                length = len(read)
            else:
                if len(read) != length:
                    raise ValueError(
                        'FASTA sequence %r was not of the '
                        'expected length (%d)' % (read.id, length))

            name = read.id.split()[0]
            fp.write(('%s%s %s\n' %
                      (name, idSpace, read.sequence)).encode('utf-8'))

        if count:
            # Print the phylip header with the number of sequences and their
            # common length.
            print('%d %d' % (count, length))
            fp.seek(0)
            write = sys.stdout.write
            while True:
                chunk = fp.read(1024)
                if chunk:
                    write(chunk.decode('utf-8'))
                else:
                    break



#!/usr/bin/env python

"""
Read protein match output produced by noninteractive-alignment-panel.py and
group it by pathogen (either virus or bacteria).

This is currently only useful when you are matching against a subject protein
database whose titles have a pathogen name in square brackets, like this:

gi|820945251|ref|YP_009137096.1| envelope glycoprotein H [Human herpesvirus 1]
gi|820945301|ref|YP_009137146.1| virion protein US10 [Human herpesvirus 1]
gi|820945229|ref|YP_009137074.1| ubiquitin E3 ligase ICP0 [Human herpesvirus 1]

In this case, those three matched subjects are from the same pathogen. This
script will gather those matches under their common "Human herpesvirus 1"
title and provides methods to print them.

Files with names in this format are provided by NCBI for their viral and
bacterial refseq protein FASTA.

The script reads *file names* from standard input, and writes to standard
output.  Alternately, you can also provide file names on the command line.

Typical usage:

  $ find . -name summary-proteins | proteins-to-pathogens.py \
        --sampleNameRegex "(Sample_\\d+)/" --html > index.html

Input files must contain lines in the following format:

0.77 47.00 48.10  5  5  74 gi|101105594| ubiquitin [Brazilian marseillevirus]
0.31 42.70 48.10 47 47 630 gi|313768007| protein BpV1_008c [Bathycoccus virus]
0.21 42.00 48.10 42 42 687 gi|313768010| protein BpV1_011c [Bathycoccus virus]
0.77 46.60 48.10  5  5  74 gi|327409793| ubiquitin [Lausannevirus]
0.33 42.70 48.10 48 48 624 gi|472342805| protein 70 [Micromonas pusilla virus]

Fields must be whitespace separated. The seven fields are:

    Coverage
    Median bit score
    Best bit score
    Read count
    HSP count
    Protein length
    Title (in the format "protein name [pathogen name]")
"""

from __future__ import print_function
import argparse
import sys
from itertools import chain

# It's not clear that the PDF backend is the right choice here, but it
# works (i.e., the generation of PNG images works fine).
import matplotlib
matplotlib.use('PDF')

# These imports are here because dark.proteins imports matplotlib.pyplot
# and we need to set the matplotlib backend before the import. So please
# don't move this import higher in this file.

from dark.proteins import ProteinGrouper


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Group proteins by the pathogen they're from.")

    parser.add_argument(
        'filenames', nargs='*', help='Sample file names to read input from.')

    parser.add_argument(
        '--sampleName',
        help=('An (optional) sample name. This is only used in producing '
              'HTML output. Should be used when all input files are for a '
              'single sample. Cannot be used with --sampleNameRegex.'))

    parser.add_argument(
        '--sampleNameRegex',
        help=('An (optional) regular expression that can be used to extract a '
              'short sample name from full sample file name.  The regular '
              'expression must have a matching group (delimited by '
              'parentheses) that captures the part of the file name that '
              'should be used as the sample name.'))

    parser.add_argument(
        '--pathogenPanelFilename',
        help=('An (optional) filename to write a pathogen-sample panel PNG '
              'image to.'))

    parser.add_argument(
        '--sampleIndexFilename',
        help=('An (optional) filename to write a sample index file to. '
              'Lines in the file will have an integer index, a space, and '
              'then the sample name. Only produced if --html is used '
              '(because the pathogen-NNN-sample-MMM.fastq are only written '
              'in that case).'))

    parser.add_argument(
        '--pathogenIndexFilename',
        help=('An (optional) filename to write a pathogen index file to. '
              'Lines in the file will have an integer index, a space, and '
              'then the pathogen name. Only produced if --html is used '
              '(because the pathogen-NNN-sample-MMM.fastq are only written '
              'in that case).'))

    parser.add_argument(
        '--html', default=False, action='store_true',
        help='If specified, output HTML instead of plain text.')

    parser.add_argument(
        '--format', default='fasta', choices=('fasta', 'fastq'),
        help=('Give the format of the sequence files written by '
              'noninteractive-alignment-panel.py when it created the '
              'summary-proteins files given on output.'))

    parser.add_argument(
        '--proteinFastaFilename', '--pff', nargs='+', action='append',
        help=('Optional filename(s) giving the name of the FASTA file(s) '
              'with the protein AA sequences with their associated pathogens '
              'in square brackets. This is the format used by NCBI for '
              'bacterial and viral reference sequence protein files. If '
              'given, the contents of this file will be used to determine how '
              'many proteins each matched pathogen has. This makes it much '
              'easier to spot significant matches (as opposed to those where, '
              'say, just one protein from a pathogen is matched).'))

    parser.add_argument(
        '--minProteinFraction', type=float, default=0.0,
        help=('The minimum fraction of proteins in a pathogen that must be '
              'matched by a particular sample in order for that pathogen to '
              'be displayed for that sample.'))

    parser.add_argument(
        '--minProteinCount', type=int, default=0,
        help=('The minimum number of proteins in a pathogen that must be '
              'matched by a particular sample in order for that pathogen to '
              'be displayed for that sample.'))

    parser.add_argument(
        '--pathogenType', default='viral', choices=('bacterial', 'viral'),
        help=('Specify the pathogen type. This option only affects the '
              'language used in HTML output.'))

    parser.add_argument(
        '--showReadLengths', default=False, action='store_true',
        help=('If specified, the HTML output (use --html to get this) will '
              'contain the lengths of all reads that match proteins for a '
              'pathogen.'))

    parser.add_argument(
        '--assetDir', default='out',
        help=('The output directory where noninteractive-alignment-panel.py '
              'puts its HTML, plots and FASTA or FASTQ files, needed for '
              'using --html.'))

    parser.add_argument(
        '--pathogenDataDir', default='pathogen-data',
        help=('The directory where per-pathogen information (e.g., collected '
              'reads across all samples) should be written.'))

    parser.add_argument(
        '--title', default='Summary of pathogens',
        help='The title to show at the top of the output.')

    parser.add_argument(
        '--preamble',
        help='Optional preamble text to show after the title.')

    parser.add_argument(
        '--titleRegex', default=None,
        help='A regex that pathogen names must match.')

    parser.add_argument(
        '--negativeTitleRegex', default=None,
        help='a regex that pathogen names must not match.')

    parser.add_argument(
        '--omitVirusLinks', default=False, action='store_true',
        help=('If specified, the HTML output (use --html to get this) for '
              'viruses will not contain links to ICTV and ViralZone. '
              'This should be used when working with viruses that do not yet '
              'have names that can be looked up.'))

    parser.add_argument(
        '--omitSampleProteinCount', default=False, action='store_true',
        help=('If specified, the HTML output (use --html to get this) for '
              'viruses will not contain counts of the number of proteins '
              'matched by each sample for a given pathogen. This should be '
              'used when working with RVDB where there are many sequences '
              'for some proteins and a sample matches many of them, leading '
              'to incorrect reporting of the number of proteins of a pathogen '
              'that are matched by samples.'))

    args = parser.parse_args()

    if args.sampleName and args.sampleNameRegex:
        print('It does not make sense to use --sampleName '
              'as well as --sampleNameRegex', file=sys.stderr)
        sys.exit(1)

    if not args.html:
        if args.sampleIndexFilename:
            print('It does not make sense to use --sampleIndexFilename '
                  'without also using --html', file=sys.stderr)
            sys.exit(1)
        if args.pathogenIndexFilename:
            print('It does not make sense to use --pathogenIndexFilename '
                  'without also using --html', file=sys.stderr)
            sys.exit(1)
        if args.omitVirusLinks:
            print('It does not make sense to use --omitVirusLinks '
                  'without also using --html', file=sys.stderr)
            sys.exit(1)
        if args.omitSampleProteinCount:
            print('It does not make sense to use --omitSampleProteinCount '
                  'without also using --html', file=sys.stderr)
            sys.exit(1)

    if args.omitVirusLinks and args.pathogenType != 'viral':
        print('The --omitVirusLinks option only makes sense with '
              '--pathogenType viral', file=sys.stderr)
        sys.exit(1)

    if args.proteinFastaFilename:
        # Flatten lists of lists that we get from using both nargs='+' and
        # action='append'. We use both because it allows people to use
        # (e.g.)  --pff on the command line either via "--pff file1 --pff
        # file2" or "--pff file1 file2", or a combination of these. That
        # way it's not necessary to remember which way you're supposed to
        # use it and you also can't be hit by the subtle problem
        # encountered in https://github.com/acorg/dark-matter/issues/453
        proteinFastaFilenames = list(chain.from_iterable(
            args.proteinFastaFilename))
    else:
        proteinFastaFilenames = None

    grouper = ProteinGrouper(assetDir=args.assetDir,
                             sampleName=args.sampleName,
                             sampleNameRegex=args.sampleNameRegex,
                             format_=args.format,
                             proteinFastaFilenames=proteinFastaFilenames,
                             saveReadLengths=args.showReadLengths,
                             titleRegex=args.titleRegex,
                             negativeTitleRegex=args.negativeTitleRegex,
                             pathogenDataDir=args.pathogenDataDir)

    if args.filenames:
        filenames = args.filenames
    else:
        filenames = (line[:-1] for line in sys.stdin)

    for filename in filenames:
        with open(filename) as fp:
            grouper.addFile(filename, fp)

    if args.html:
        print(grouper.toHTML(
            args.pathogenPanelFilename,
            minProteinFraction=args.minProteinFraction,
            minProteinCount=args.minProteinCount,
            pathogenType=args.pathogenType,
            title=args.title, preamble=args.preamble,
            sampleIndexFilename=args.sampleIndexFilename,
            pathogenIndexFilename=args.pathogenIndexFilename,
            omitVirusLinks=args.omitVirusLinks,
            omitSampleProteinCount=args.omitSampleProteinCount))
    else:
        print(grouper.toStr())



#!/usr/bin/env python

from __future__ import print_function

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            'Given FASTA on stdin, write the sequences to stdout.'))

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)

    for read in reads:
        print(read.sequence)



from six import PY3
from hashlib import md5
import sqlite3
import os

from Bio import SeqIO, bgzf
from pyfaidx import Fasta

from dark.reads import Reads, DNARead
from dark.utils import asHandle


def fastaToList(fastaFilename):
    return list(SeqIO.parse(fastaFilename, 'fasta'))


def dedupFasta(reads):
    """
    Remove sequence duplicates (based on sequence) from FASTA.

    @param reads: a C{dark.reads.Reads} instance.
    @return: a generator of C{dark.reads.Read} instances with no duplicates.
    """
    seen = set()
    add = seen.add
    for read in reads:
        hash_ = md5(read.sequence.encode('UTF-8')).digest()
        if hash_ not in seen:
            add(hash_)
            yield read


def dePrefixAndSuffixFasta(sequences):
    """
    sequences: an iterator producing Bio.Seq sequences.

    return: a generator of sequences with no duplicates and no fully contained
        subsequences.
    """
    sequences = sorted(sequences, key=lambda s: len(s.seq), reverse=True)
    seen = set()
    for s in sequences:
        thisSeq = str(s.seq)
        thisHash = md5(thisSeq.encode('UTF-8')).digest()
        if thisHash not in seen:
            # Add prefixes.
            newHash = md5()
            for nucl in thisSeq:
                newHash.update(nucl.encode('UTF-8'))
                seen.add(newHash.digest())
            # Add suffixes.
            for start in range(len(thisSeq) - 1):
                seen.add(md5(thisSeq[start + 1:].encode('UTF-8')).digest())
            yield s


def fastaSubtract(fastaFiles):
    """
    Given a list of open file descriptors, each with FASTA content,
    remove the reads found in the 2nd, 3rd, etc files from the first file
    in the list.

    @param fastaFiles: a C{list} of FASTA filenames.
    @raises IndexError: if passed an empty list.
    @return: An iterator producing C{Bio.SeqRecord} instances suitable for
        writing to a file using C{Bio.SeqIO.write}.

    """
    reads = {}
    firstFile = fastaFiles.pop(0)
    for seq in SeqIO.parse(firstFile, 'fasta'):
        reads[seq.id] = seq

    for fastaFile in fastaFiles:
        for seq in SeqIO.parse(fastaFile, 'fasta'):
            # Make sure that reads with the same id have the same sequence.
            if seq.id in reads:
                assert str(seq.seq) == str(reads[seq.id].seq)
            reads.pop(seq.id, None)

    return iter(reads.values())


class FastaReads(Reads):
    """
    Subclass of L{dark.reads.Reads} providing access to FASTA reads.

    @param _files: Either a single C{str} file name or file handle, or a
        C{list} of C{str} file names and/or file handles. Each file or file
        handle must contain sequences in FASTA format.
    @param readClass: The class of read that should be yielded by iter.
    @param upperCase: If C{True}, read sequences will be converted to upper
        case.
    """
    def __init__(self, _files, readClass=DNARead, upperCase=False):
        self._files = _files if isinstance(_files, (list, tuple)) else [_files]
        self._readClass = readClass
        # TODO: It would be better if upperCase were an argument that could
        # be passed to Reads.__init__ and that could do the uppercasing in
        # its add method (as opposed to using it below in our iter method).
        # In that case, in the iter of this class we'd call self.add on
        # each of the sequences coming from self._file. Or, if we'd already
        # read the file we'd return Reads.iter(self) to re-iterate over the
        # sequences already added from the file.
        self._upperCase = upperCase
        if PY3:
            super().__init__()
        else:
            Reads.__init__(self)

    def iter(self):
        """
        Iterate over the sequences in the files in self.files_, yielding each
        as an instance of the desired read class.
        """
        count = 0
        for _file in self._files:
            with asHandle(_file) as fp:
                # Duplicate some code here so as not to test
                # self._upperCase in the loop.
                if self._upperCase:
                    for seq in SeqIO.parse(fp, 'fasta'):
                        read = self._readClass(seq.description,
                                               str(seq.seq.upper()))
                        yield read
                        count += 1
                else:
                    for seq in SeqIO.parse(fp, 'fasta'):
                        read = self._readClass(seq.description, str(seq.seq))
                        yield read
                        count += 1


class FastaFaiReads(Reads):
    """
    Subclass of L{dark.reads.Reads} that provides dictionary-like access to
    FASTA reads using the pyfaidx module.

    @param filename: The C{str} name of a file containing FASTA. This may be an
        uncompressed file or one that has been compressed with bgzip (from
        the samtools suite).
        # The pyfaidx module will create an identically named FASTA index file
        # with an additional ".fai" extension. An .fai file can also be made
        # (if desired) with 'samtools faidx' or the 'faidx' command installed
        # by pyfaidx. If you use the latter, you can preserve full sequence
        # titles via '-e "lambda x: x".
    @param readClass: The class of read that should be yielded by iter.
    @param upperCase: If C{True}, read sequences will be converted to upper
        case.
    """
    def __init__(self, filename, readClass=DNARead, upperCase=False):
        self._filename = filename
        self._fasta = Fasta(filename)
        self._readClass = readClass
        # TODO: It would be better if upperCase were an argument that could
        # be passed to Reads.__init__ and that could do the uppercasing in
        # its add method (as opposed to using it below in our iter method).
        # In that case, in the iter of this class we'd call self.add on
        # each of the sequences coming from self._file. Or, if we'd already
        # read the file we'd return Reads.iter(self) to re-iterate over the
        # sequences already added from the file.
        self._upperCase = upperCase
        if PY3:
            super().__init__()
        else:
            Reads.__init__(self)

    def iter(self):
        """
        Iterate over the sequences in the files in self.files_, yielding each
        as an instance of the desired read class.
        """
        if self._upperCase:
            for id_ in self._fasta:
                yield self._readClass(id_, str(self._fasta[id_]).upper())
        else:
            for id_ in self._fasta:
                yield self._readClass(id_, str(self._fasta[id_]))

    def __getitem__(self, id_):
        return self._readClass(str(id_), str(self._fasta[id_]))


def combineReads(filename, sequences, readClass=DNARead,
                 upperCase=False, idPrefix='command-line-read-'):
    """
    Combine FASTA reads from a file and/or sequence strings.

    @param filename: A C{str} file name containing FASTA reads.
    @param sequences: A C{list} of C{str} sequences. If a sequence
        contains spaces, the last field (after splitting on spaces) will be
        used as the sequence and the first fields will be used as the sequence
        id.
    @param readClass: The class of the individual reads.
    @param upperCase: If C{True}, reads will be converted to upper case.
    @param idPrefix: The C{str} prefix that will be used for the id of the
        sequences in C{sequences} that do not have an id specified. A trailing
        sequence number will be appended to this prefix. Note that
        'command-line-read-', the default id prefix, could collide with ids in
        the FASTA file, if given. So output might be ambiguous. That's why we
        allow the caller to specify a custom prefix.
    @return: A C{FastaReads} instance.
    """
    # Read sequences from a FASTA file, if given.
    if filename:
        reads = FastaReads(filename, readClass=readClass, upperCase=upperCase)
    else:
        reads = Reads()

    # Add any individually specified subject sequences.
    if sequences:
        for count, sequence in enumerate(sequences, start=1):
            # Try splitting the sequence on its last space and using the
            # first part of the split as the read id. If there's no space,
            # assign a generic id.
            parts = sequence.rsplit(' ', 1)
            if len(parts) == 2:
                readId, sequence = parts
            else:
                readId = '%s%d' % (idPrefix, count)
            if upperCase:
                sequence = sequence.upper()
            read = readClass(readId, sequence)
            reads.add(read)

    return reads


class SqliteIndex(object):
    """
    Create an Sqlite3 database holding FASTA sequence ids, file names, and
    offsets for fast random dictionary-like access.

    @param dbFilename: A C{str} file name containing an sqlite3 database. If
        the file does not exist it will be created. The special string
        ":memory:" can be used to create an in-memory database.
    @param readClass: The class of read that should be returned by __getitem__.
    @param fastaDirectory: A C{str} directory where the indexed FASTA files
        can be found. If provided, this directory is only used by __getitem__,
        which will combine it with the basename of the files given to
        C{addFile} to locate the FASTA.
    """
    def __init__(self, dbFilename, readClass=DNARead, fastaDirectory=None):
        self._readClass = readClass
        self._fastaDirectory = fastaDirectory
        creating = dbFilename == ':memory:' or not os.path.exists(dbFilename)
        self._connection = sqlite3.connect(dbFilename)
        if creating:
            # Create a new database.
            cur = self._connection.cursor()
            cur.executescript('''
                CREATE TABLE files (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    name VARCHAR UNIQUE
                );

                CREATE TABLE sequences (
                    id VARCHAR UNIQUE PRIMARY KEY,
                    fileNumber INTEGER,
                    offset INTEGER
                );
            ''')
            self._connection.commit()

    def _getFilename(self, fileNumber):
        """
        Given a file number, get its name (if any).

        @param fileNumber: An C{int} file number.
        @return: A C{str} file name or C{None} if a file with that number
            has not been added.
        """
        cur = self._connection.cursor()
        cur.execute('SELECT name FROM files WHERE id = ?', (fileNumber,))
        row = cur.fetchone()
        if row is None:
            return None
        else:
            return row[0]

    def _getFileNumber(self, filename):
        """
        Given a file name, get its file number (if any).

        @param filename: A C{str} file name.
        @return: An C{int} file number or C{None} if no file with that name
            has been added.
        """
        cur = self._connection.cursor()
        cur.execute('SELECT id FROM files WHERE name = ?', (filename,))
        row = cur.fetchone()
        if row is None:
            return None
        else:
            return row[0]

    def _addFilename(self, filename):
        """
        Add a new file name.

        @param filename: A C{str} file name.
        @raise ValueError: If a file with this name has already been added.
        @return: The C{int} id of the newly added file.
        """
        cur = self._connection.cursor()
        try:
            cur.execute('INSERT INTO files(name) VALUES (?)', (filename,))
        except sqlite3.IntegrityError as e:
            if str(e).find('UNIQUE constraint failed') > -1:
                raise ValueError('Duplicate file name: %r' % filename)
            else:
                raise
        else:
            fileNumber = cur.lastrowid
            self._connection.commit()
            return fileNumber

    def addFile(self, filename):
        """
        Add a new FASTA file of sequences.

        @param filename: A C{str} file name, with the file in FASTA format.
            This file must (obviously) exist at indexing time. When __getitem__
            is used to access sequences, it is possible to provide a
            C{fastaDirectory} argument to our C{__init__} to indicate the
            directory containing the original FASTA files, in which case the
            basename of the file here provided in C{filename} is used to find
            the file in the given directory. This allows the construction of a
            sqlite database from the shell in one directory and its use
            programmatically from another directory.
        @raise ValueError: If a file with this name has already been added or
            if the file contains a sequence whose id has already been seen.
        @return: The C{int} number of sequences added from the file.
        """
        endswith = filename.lower().endswith
        if endswith('.bgz') or endswith('.gz'):
            useBgzf = True
        elif endswith('.bz2'):
            raise ValueError(
                'Compressed FASTA is only supported in BGZF format. Use '
                'bgzip to compresss your FASTA.')
        else:
            useBgzf = False

        fileNumber = self._addFilename(filename)
        connection = self._connection
        count = 0
        try:
            with connection:
                if useBgzf:
                    try:
                        fp = bgzf.open(filename, 'rb')
                    except ValueError as e:
                        if str(e).find('BGZF') > -1:
                            raise ValueError(
                                'Compressed FASTA is only supported in BGZF '
                                'format. Use the samtools bgzip utility '
                                '(instead of gzip) to compresss your FASTA.')
                        else:
                            raise
                    else:
                        try:
                            for line in fp:
                                if line[0] == '>':
                                    count += 1
                                    id_ = line[1:].rstrip(' \t\n\r')
                                    connection.execute(
                                        'INSERT INTO sequences(id, '
                                        'fileNumber, offset) VALUES (?, ?, ?)',
                                        (id_, fileNumber, fp.tell()))
                        finally:
                            fp.close()
                else:
                    with open(filename) as fp:
                        offset = 0
                        for line in fp:
                            offset += len(line)
                            if line[0] == '>':
                                count += 1
                                id_ = line[1:].rstrip(' \t\n\r')
                                connection.execute(
                                    'INSERT INTO sequences(id, fileNumber, '
                                    'offset) VALUES (?, ?, ?)',
                                    (id_, fileNumber, offset))
        except sqlite3.IntegrityError as e:
            if str(e).find('UNIQUE constraint failed') > -1:
                original = self._find(id_)
                if original is None:
                    # The id must have appeared twice in the current file,
                    # because we could not look it up in the database
                    # (i.e., it was INSERTed but not committed).
                    raise ValueError(
                        "FASTA sequence id '%s' found twice in file '%s'." %
                        (id_, filename))
                else:
                    origFilename, _ = original
                    raise ValueError(
                        "FASTA sequence id '%s', found in file '%s', was "
                        "previously added from file '%s'." %
                        (id_, filename, origFilename))
            else:
                raise
        else:
            return count

    def _find(self, id_):
        """
        Find the filename and offset of a sequence, given its id.

        @param id_: A C{str} sequence id.
        @return: A 2-tuple, containing the C{str} file name and C{int} offset
            within that file of the sequence.
        """
        cur = self._connection.cursor()
        cur.execute(
            'SELECT fileNumber, offset FROM sequences WHERE id = ?', (id_,))
        row = cur.fetchone()
        if row is None:
            return None
        else:
            return self._getFilename(row[0]), row[1]

    def __getitem__(self, id_):
        """
        Return a read, given its id.

        @param id_: A C{str} sequence id.
        @raise KeyError: If C{id_} is not a known sequence.
        @return: A read of our read class.
        """
        location = self._find(id_)
        if location is None:
            raise KeyError('Unknown sequence: %r' % id_)
        else:
            filename, offset = location
            # If a FASTA directory was provided, look for the FASTA files
            # there, otherwise use the filename that was given to addFile.
            if self._fastaDirectory:
                filename = os.path.join(self._fastaDirectory,
                                        os.path.basename(filename))

            endswith = filename.lower().endswith
            if endswith('.bgz') or endswith('.gz'):
                opener = bgzf.open
            else:
                opener = open

            sequence = ''
            with opener(filename) as fp:
                fp.seek(offset)
                while True:
                    line = fp.readline()
                    if not line:
                        # EOF
                        break
                    elif line[0] == '>':
                        # We found the next sequence identifier.
                        break
                    else:
                        sequence += line.rstrip('\n\r')

            return self._readClass(id_, sequence)

    def close(self):
        self._connection.close()
        self._connection = None



#!/usr/bin/env python

from __future__ import print_function

import sys
import argparse

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Find sequences that have a specific property (as detected '
                 'by the "relabel" function in the file specified by '
                 '--relabelFunctionFile) and print a TAB-separated list of '
                 'renamings to standard output.'))

parser.add_argument(
    '--relabelFunctionFile', required=True,
    help=('A file containing a Python function named "relabel" that takes a '
          'list of reads and returns a dictionary of renamings.'))

addFASTACommandLineOptions(parser)
args = parser.parse_args()
reads = parseFASTACommandLineOptions(args)

relabel = origRelabel = object()

try:
    exec(open(args.relabelFunctionFile).read())
except Exception as e:
    print('Could not execute Python in %s: %s' % (args.relabelFunctionFile, e),
          file=sys.stderr)
    sys.exit(1)

if relabel is origRelabel:
    print('The Python file %s did not define a function called "relabel".' %
          args.relabelFunctionFile, file=sys.stderr)
    sys.exit(2)

for old, new in relabel(reads).items():
    print('%s\t%s' % (old, new))



from unittest import TestCase
from six import assertRaisesRegex

from dark.cigar import CINS, CDEL, CMATCH, CEQUAL, CDIFF, dna2cigar


class TestConstants(TestCase):
    """
    Test constants in dark.cigar
    """
    def testOperations(self):
        """
        Make sure the CIGAR operations have the expected one-letter codes.
        """
        self.assertEqual('I', CINS)
        self.assertEqual('D', CDEL)
        self.assertEqual('M', CMATCH)
        self.assertEqual('=', CEQUAL)
        self.assertEqual('X', CDIFF)


class TestDNA2CIGAR(TestCase):
    """
    Test the dna2cigar function.
    """
    def testUnequalStrings(self):
        """
        If two unequal strings are passed, a ValueError must be raised.
        """
        error = r"^Sequences 'hey' and 'there' of unequal length \(3 != 5\)\.$"
        assertRaisesRegex(self, ValueError, error, dna2cigar, 'hey', 'there')

    def testEmptyStrings(self):
        """
        If two empty strings are passed, a ValueError must be raised.
        """
        error = r"^Two sequences of zero length were passed\.$"
        assertRaisesRegex(self, ValueError, error, dna2cigar, '', '')

    def testEqualStringsLength1(self):
        """
        If two equal strings of length 1 are passed, the correct CIGAR string
        must be returned.
        """
        self.assertEqual('1=', dna2cigar('A', 'A'))

    def testEqualStringsLength1MismatchedCase(self):
        """
        If two strings of length 1 are passed and the're the same but for
        case, the CIGAR string must indicate that they do not match.
        """
        self.assertEqual('1X', dna2cigar('A', 'a'))

    def testEqualStringsLength1Concise(self):
        """
        If two equal strings of length 1 are passed and the concise argument
        is True, the correct CIGAR string must be returned.
        """
        self.assertEqual('1M', dna2cigar('A', 'A', concise=True))

    def testUnequalStringsLength1(self):
        """
        If two unequal strings of length 1 are passed, the correct CIGAR
        string must be returned.
        """
        self.assertEqual('1X', dna2cigar('A', 'G'))

    def testUnequalStringsLength1Concise(self):
        """
        If two unequal strings of length 1 are passed and the concise argument
        is True, the correct CIGAR string must be returned.
        """
        self.assertEqual('1M', dna2cigar('A', 'G', concise=True))

    def testEqualStringsLength2(self):
        """
        If two equal strings of length 2 are passed, the correct CIGAR string
        must be returned.
        """
        self.assertEqual('2=', dna2cigar('AT', 'AT'))

    def testEqualStringsLength2Concise(self):
        """
        If two equal strings of length 2 are passed and the concise argument
        is True, the correct CIGAR string must be returned.
        """
        self.assertEqual('2M', dna2cigar('AT', 'AT', concise=True))

    def testUnequalStringsLength2(self):
        """
        If two unequal strings of length 2 are passed, the correct CIGAR
        string must be returned.
        """
        self.assertEqual('2X', dna2cigar('AT', 'GC'))

    def testUnequalStringsLength2Concise(self):
        """
        If two unequal strings of length 2 are passed and the concise argument
        is True, the correct CIGAR string must be returned.
        """
        self.assertEqual('2M', dna2cigar('AT', 'GC', concise=True))

    def testMixedEqualityStringsLength2(self):
        """
        If two strings of length 2 are passed and they match in just one
        place, the correct CIGAR string must be returned.
        """
        self.assertEqual('1=1X', dna2cigar('AG', 'AT'))

    def testMixedEqualityStringsLength2Concise(self):
        """
        If two equal strings of length 2 are passed and the concise argument
        is True, the correct CIGAR string must be returned.
        """
        self.assertEqual('2M', dna2cigar('AG', 'AT', concise=True))

    def testMixedEqualityStringsLength2Alt(self):
        """
        If two strings of length 2 are passed and they match in just one
        place, the correct CIGAR string must be returned.
        """
        self.assertEqual('1X1=', dna2cigar('GA', 'TA'))

    def testMixedEqualityStringsLength2ConciseAlt(self):
        """
        If two equal strings of length 2 are passed and the concise argument
        is True, the correct CIGAR string must be returned.
        """
        self.assertEqual('2M', dna2cigar('GA', 'TA', concise=True))

    def testMixedMatch(self):
        """
        If two strings with matching and non-matching regions must result
        in the expected CIGAR string.
        """
        self.assertEqual('3=4X5=2X',
                         dna2cigar('ACGTTTTCCCTTGG',
                                   'ACGAAAACCCTTTT'))

    def testMixedMatchConcise(self):
        """
        If two strings with matching and non-matching regions must result
        in the expected CIGAR string when the concise argument is True.
        """
        self.assertEqual('14M',
                         dna2cigar('ACGTTTTCCCTTGG',
                                   'ACGAAAACCCTTTT', concise=True))
